/**
   @author Shin'ichiro Nakaoka
*/

#include "Archive.h"
#include "Item.h"
#include "App.h"
#include "AppConfig.h"
#include <cnoid/Referenced>
#include <map>
#include <vector>
#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <boost/signals.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;
using namespace boost;
using namespace cnoid;


namespace {
#ifdef WIN32
    const char* HOME = "HOMEPATH";
#else
    const char* HOME = "HOME";
#endif
    regex regVariable("^\\$\\{(\\w+)\\}");
    typedef map<ItemPtr, int> ItemToIdMap;
    const char* const APP_CONFIG_STR_PATH_VARIABLES = "PathVariables";
}


namespace cnoid {
    
    class ArchiveSharedData : public Referenced
    {
    public:
        filesystem::path projectTopDir;
        string homePathString;
        string programTopDirString;
        //string shareDirString;
        regex regProject;
        regex regHome;
        regex regProgramTopDir;
        //regex regShare;

        vector<Item*> idToItems;
        ItemToIdMap itemToIds;

        Item* currentParentItem;

        boost::signal<void()> postProcesses;
    };

    string getAbsPathString(string orgpath)
    {
        filesystem::path absPath(filesystem::system_complete(filesystem::path(orgpath)));
        string rootPath = boost::to_upper_copy(absPath.root_path().string());
        string absPathStr(absPath.string());
        if(!rootPath.empty())
        {
            string::size_type pos = absPathStr.find(absPath.root_path().string(), 0);

            if(pos != string::npos)
            {
                absPathStr.replace(pos, absPath.root_path().string().length(), rootPath);
            }
        }
        return absPathStr;
    }


    boost::filesystem::path
    to_relative_path(boost::filesystem::path const path, boost::filesystem::path const base)
    {
        if ( !path.is_complete() || !base.is_complete() )
            return path;

        if ( path.root_path() != base.root_path() )
            return path;
            
        boost::filesystem::path::const_iterator path_it = path.begin();
        boost::filesystem::path::const_iterator base_it = base.begin();
        while ( path_it != path.end() && base_it != base.end() ) {
            if (*path_it != *base_it) break;
            ++path_it; ++base_it;
        }
        boost::filesystem::path result;
        for (; base_it != base.end(); ++base_it) {
            result /= "..";
        }
        for (; path_it != path.end(); ++path_it) {
            result /= *path_it;
        }
        return result;
    }

    vector<string> split_string(const string& str,
                                const string& delim)
    {
        size_t offset = 0;
        size_t pos = 0;
        vector<string> result;
        
        while((pos = str.find_first_of(delim, offset)) != str.npos)
        {
            if(offset != pos)
            {
                result.push_back(string(str.begin() + offset, str.begin() + pos));
            }
            offset = pos + 1;
        }
        result.push_back(string(str.begin() + offset, str.end()));
        return result;
    }

    bool replacePathToVariable(const std::string& origin,
                               std::string& result)
    {
        filesystem::path originPath(getAbsPathString(origin));

        YamlMapping* map = AppConfig::archive()->openMapping(APP_CONFIG_STR_PATH_VARIABLES);
        for(YamlMapping::const_iterator ite = map->begin();
            ite != map->end();
            ++ite)
        {
            vector<string> vpath = split_string(ite->second->toString(), ";");
            for(vector<string>::iterator ipath = vpath.begin();
                ipath != vpath.end();
                ++ipath)
            {
                filesystem::path vPath(getAbsPathString(*ipath));
                if ( !originPath.is_complete() || !vPath.is_complete() )
                    continue;

                if ( originPath.root_path() != vPath.root_path() )
                    continue;
                
                boost::filesystem::path::const_iterator o_it = originPath.begin();
                boost::filesystem::path::const_iterator v_it = vPath.begin();
                while ( o_it != originPath.end() && v_it != vPath.end() ) {
                    if (*o_it != *v_it) break;
                    ++o_it; ++v_it;
                }

                if(v_it == vPath.end())
                {
                    boost::filesystem::path res_path;
                    for (; o_it != originPath.end(); ++o_it) {
                        res_path /= *o_it;
                    }
                    result = string("${") + ite->first + string("}/") + res_path.string();
                    return true;
                }
            }
        }
        return false;
    }

    std::string replaceVariableToPath(const std::string& origin,
                                      const std::string varname,
                                      size_t pos,
                                      size_t len)
    {
        YamlMapping* map = AppConfig::archive()->openMapping(APP_CONFIG_STR_PATH_VARIABLES);
        for(YamlMapping::const_iterator ite = map->begin();
            ite != map->end();
            ++ite)
        {
            if(ite->first == varname)
            {
                vector<string> vpath = split_string(ite->second->toString(), ";");
                for(vector<string>::iterator ipath = vpath.begin();
                    ipath != vpath.end();
                    ++ipath)
                {
                    string ret = origin;
                    ret.replace(pos, len, *ipath);
                    filesystem::path reppath(ret);
                    filesystem::file_status fstatus = filesystem::status(reppath);
                    if(filesystem::is_directory(fstatus) || filesystem::exists(fstatus))
                    {
                        return ret;
                    }
                }
            }
        }
        
        string ret = origin;
        ret.replace(pos, len, ".");
        return ret;
    }
}


ArchivePtr Archive::invalidArchive()
{
    static ArchivePtr invalidArchive_ = new Archive();
    invalidArchive_->type_ = YAML_NONE;
    return invalidArchive_;
}


Archive::Archive()
{

}


Archive::Archive(int line, int column)
    : YamlMapping(line, column)
{

}


Archive::~Archive()
{

}


void Archive::initSharedInfo(const std::string& projectFile)
{
    shared = new ArchiveSharedData;

    filesystem::path projectPath(filesystem::system_complete(filesystem::path(projectFile)));
    shared->projectTopDir = projectPath.branch_path();
    shared->regProject = regex(string("^") + shared->projectTopDir.string() + "(/(.+)|())$");
    char* homePath = getenv(HOME);
    if(homePath){
        shared->homePathString = homePath;
    } else {
        shared->homePathString = "UnkownHomePath";
    }
    shared->regHome = regex(string("^") + shared->homePathString + "/(.+)$");
    shared->programTopDirString = App::topDirectory();
    shared->regProgramTopDir = regex(string("^") + shared->programTopDirString + "/(.+)$");
    //regex regShare(string("^") + shareDirString + "/(.+)$");
    shared->currentParentItem = 0;
}


void Archive::inheritSharedInfoFrom(ArchivePtr archive)
{
    shared = archive->shared;
}


void Archive::addPostProcess(const boost::function<void()>& func) const
{
    if(shared){
        shared->postProcesses.connect(func);
    }
}


void Archive::callPostProcesses()
{
    if(shared){
        shared->postProcesses();
    }
}


Archive* Archive::findSubArchive(const std::string& name)
{
    YamlMapping* mapping = findMapping(name);
    if(mapping->isValid()){
        Archive* archive = dynamic_cast<Archive*>(mapping);
        if(archive){
            return archive;
        }
    }

    return invalidArchive().get();
}

bool Archive::readRelocatablePath(const std::string& key, std::string& out_value) const
{
    string orgPathString;

    if(read(key, orgPathString)){

        boost::smatch results;

        // expand variables in the path 
        if(regex_search(orgPathString, results, regVariable)){
            
            string varname(results.str(1));
            const size_t pos = results.position(); // position(0)
            const size_t len = results.length();   // length(0)

            if(varname == "PROGRAM_TOP"){
                orgPathString.replace(pos, len, shared->programTopDirString);
            } else if(varname == "HOME"){
                orgPathString.replace(pos, len, shared->homePathString);
            }
            else
            {
                orgPathString = replaceVariableToPath(orgPathString, varname, pos, len);
            }

            
            /*
              else if(varname == "SHARE_DIR"){
                orgPathString.replace(pos, len, shareDirString);
            }
            */
        }

        filesystem::path orgPath(orgPathString);

        if(orgPath.is_complete()){
            out_value = orgPath.file_string();
        } else {
            out_value = (shared->projectTopDir / orgPath).file_string();
        } 
            
        return true;
    }

    return false;
}


void Archive::writeRelocatablePath(const std::string& key, const std::string& orgpath)
{
    string absPathStr = getAbsPathString(orgpath);

    boost::smatch results;
    string ret;

    if(regex_match(absPathStr, results, shared->regProject)){
        string relpath = results.str(2);
        if(relpath.empty()){
            relpath = ".";
        }
        write(key, relpath, YAML_DOUBLE_QUOTED);
    } else if(regex_match(absPathStr, results, shared->regProgramTopDir)){
        write(key, string("${PROGRAM_TOP}/") + results.str(1), YAML_DOUBLE_QUOTED);
    }
    else if(replacePathToVariable(absPathStr, ret))
    {
        write(key, ret, YAML_DOUBLE_QUOTED);
    }
    /*
    else if(regex_match(absPathStr, results, shared->regHome)){
        write(key, string("${HOME}/") + results.str(1), YAML_DOUBLE_QUOTED);
        }*/
    else {
        boost::filesystem::path relative_path = to_relative_path(absPathStr, shared->projectTopDir);
        string relativePathStr(relative_path.string());
        write(key, relativePathStr, YAML_DOUBLE_QUOTED);
    }
        
    /*
    } else if(regex_match(absPathStr, results, regShare)){
        write(key, string("${SHARE_DIR}/") + results.str(1));
    }
    */
}

void Archive::registerItemId(Item* item, int id)
{
    if(shared){
        if(id >= (signed)shared->idToItems.size()){
            shared->idToItems.resize(id + 1);
        }
        shared->idToItems[id] = item;
        shared->itemToIds[item] = id;
    }
}


/**
   @return -1 if item does not belong to the archive
*/
//int Archive::getItemId(Item* item) const
int Archive::getItemId(ItemPtr item) const
{
    if(shared && item){
        ItemToIdMap::const_iterator p = shared->itemToIds.find(item);
        if(p != shared->itemToIds.end()){
            return p->second;
        }
    }
    return -1;
}


void Archive::writeItemId(const std::string& key, ItemPtr item)
{
    if(item){
        int id = getItemId(item);
        if(id >= 0){
            write(key, id);
        }
    }
}


Item* Archive::findItem(int id) const
{
    if(shared){
        if(id >= 0 && id < (signed)shared->idToItems.size()){
            return shared->idToItems[id];
        }
    }
    return 0;
}


Item* Archive::currentParentItem() const
{
    if(shared){
        return shared->currentParentItem;
    }
    return 0;
}


void Archive::setCurrentParentItem(Item* parentItem)
{
    if(shared){
        shared->currentParentItem = parentItem;
    }
}
