/**
   @author Shin'ichiro Nakaoka
*/

#include "PluginManager.h"
#include "Plugin.h"
#include "ExtensionManager.h"
#include "MenuManager.h"
#include "MessageView.h"
#include "DescriptionDialog.h"
#include "App.h"
#include "LazyCaller.h"
#include <cnoid/Config>
#include <QLibrary>
#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/regex.hpp>
#include <boost/tokenizer.hpp>
#include <boost/format.hpp>
#include <boost/bind.hpp>
#include <vector>
#include <map>

#include <iostream>

using namespace std;
using namespace boost;
using namespace cnoid;


#ifdef Q_OS_WIN32
static const string DLL_PREFIX = "";
static const string DLL_SUFFIX = "dll";
static const char* PATH_DELIMITER = ";";
# ifdef CNOID_DEBUG
static const string DEBUG_SUFFIX = "d";
# else
static const string DEBUG_SUFFIX = "";
# endif
#else
# ifdef Q_OS_MAC
static const string DLL_PREFIX = "";
static const string DLL_SUFFIX = "dylib";
static const char* PATH_DELIMITER = ":";
static const string DEBUG_SUFFIX = "";
# else
static const string DLL_PREFIX = "lib";
static const string DLL_SUFFIX = "so";
static const char* PATH_DELIMITER = ":";
static const string DEBUG_SUFFIX = "";
# endif
#endif


namespace cnoid {

    class PluginManagerImpl
    {
    public:

        PluginManagerImpl();
        ~PluginManagerImpl();

        MessageView* mv;

        regex pluginNamePattern;
        regex hiddenFileDirPattern;

        struct PluginInfo {
            PluginInfo(){
                plugin = 0;
                status = PluginManager::NOT_LOADED;
                aboutMenuItem = 0;
                aboutDialog = 0;
            }
            QLibrary dll;
            std::string pathString;
            Plugin* plugin;
            std::string name;
            std::list<std::string> dependencies;
            int status;
            
            QAction* aboutMenuItem;
            DescriptionDialog* aboutDialog;
        };
        typedef shared_ptr<PluginInfo> PluginInfoPtr;

        typedef vector<PluginInfoPtr> PluginInfoArray;
        PluginInfoArray pluginInfoList;

        typedef map<std::string, PluginInfoPtr> PluginMap;
        PluginMap nameToActivePluginInfoMap;
        PluginMap pathToPluginInfoMap;

        typedef multimap<std::string, std::string> MultiNameMap;
        MultiNameMap oldNameToCurrentPluginNameMap;

        vector<int> pluginsToUnload;
        LazyCaller unloadingRequest;
    
        void clearUnusedPlugins();
        void scanPluginFilesInDefaultPath(const std::string& pathList);
        void scanPluginFilesInDirectoyOfExecFile();
        void scanPluginFiles(const std::string& pathString);
        void loadAllPlugins();
        void unloadAllPlugins();
        bool loadPlugin(int index);
        void onAboutDialogTriggered(PluginInfo* info);
        const char* guessActualPluginName(const std::string& name);
        bool unloadPlugin(int index);
        void onUnloadingRequest();
    };
}


PluginManager* PluginManager::instance()
{
    static PluginManager* pluginManager = new PluginManager();
    return pluginManager;
}


PluginManager::PluginManager()
{
    impl = new PluginManagerImpl();
}


PluginManagerImpl::PluginManagerImpl()
    : mv(MessageView::mainInstance())
{
    pluginNamePattern = DLL_PREFIX + "Cnoid.+Plugin" + DEBUG_SUFFIX + "\\." + DLL_SUFFIX;
    hiddenFileDirPattern = "^\\..+";

    // for the base module
    PluginInfoPtr info(new PluginInfo);
    info->name = "Base";
    nameToActivePluginInfoMap.insert(make_pair("Base", info));

    unloadingRequest.set(bind(&PluginManagerImpl::onUnloadingRequest, this), IDLE_PRIORITY_LOW);
}


PluginManager::~PluginManager()
{
    delete impl;
}


PluginManagerImpl::~PluginManagerImpl()
{
    unloadAllPlugins();
}


int PluginManager::numPlugins()
{
    return impl->pluginInfoList.size();
}


const std::string& PluginManager::pluginPath(int index)
{
    return impl->pluginInfoList[index]->pathString;
}


const std::string& PluginManager::pluginName(int index)
{
    return impl->pluginInfoList[index]->name;
}


int PluginManager::pluginStatus(int index)
{
    return impl->pluginInfoList[index]->status;
}


/**
   This function scans plugin files in the path list.
   @param pathList List of the directories to scan, which is specified with the same format
   as the PATH environment varibale.
*/
void PluginManager::scanPluginFilesInPathList(const std::string& pathList)
{
    impl->scanPluginFilesInDefaultPath(pathList);
}


void PluginManagerImpl::scanPluginFilesInDefaultPath(const std::string& pathList)
{
    char_separator<char> sep(PATH_DELIMITER);
    tokenizer< char_separator<char> > paths(pathList, sep);
    tokenizer< char_separator<char> >::iterator p;
    for(p = paths.begin(); p != paths.end(); ++p){
        const string& path = *p;
        scanPluginFiles(path);
    }
}


void PluginManager::scanPluginFilesInDirectoyOfExecFile()
{
  impl->scanPluginFilesInDirectoyOfExecFile();
}


void PluginManagerImpl::scanPluginFilesInDirectoyOfExecFile()
{
    string pluginDirectory(
        str(boost::format("%1%/%2%")
            % App::topDirectory() % CNOID_PLUGIN_SUBDIR));
    
    scanPluginFiles(pluginDirectory);
} 


void PluginManager::scanPluginFiles(const std::string& pathString)
{
    impl->scanPluginFiles(pathString);
}


void PluginManagerImpl::scanPluginFiles(const std::string& pathString)
{
    filesystem::path pluginPath(pathString);

    if(filesystem::exists(pluginPath)){

        if(filesystem::is_directory(pluginPath)){
            filesystem::directory_iterator end;
            for(filesystem::directory_iterator it(pluginPath); it != end; ++it){
                const filesystem::path& filepath = *it;
                if(!regex_match(filepath.leaf(), hiddenFileDirPattern)){
                    scanPluginFiles(filepath.string());
                }
            }
        } else {
            
            if(regex_match(pluginPath.leaf(), pluginNamePattern)){

                PluginMap::iterator p = pathToPluginInfoMap.find(pathString);

                if(p == pathToPluginInfoMap.end()){
                    PluginInfoPtr info(new PluginInfo);
                    info->pathString = pathString;
                    pluginInfoList.push_back(info);
                    pathToPluginInfoMap[pathString] = info;
                }
            }
        }
    }
}


void PluginManager::clearUnusedPlugins()
{
    impl->clearUnusedPlugins();
}


void PluginManagerImpl::clearUnusedPlugins()
{
    PluginInfoArray oldList = pluginInfoList;
    pluginInfoList.clear();

    for(size_t i=0; i < oldList.size(); ++i){
        PluginInfoPtr info = oldList[i];
        if(info->status == PluginManager::ACTIVE){
            pluginInfoList.push_back(info);
        } else {
            pathToPluginInfoMap.erase(info->pathString);
        }
    }
}


void PluginManager::loadAllPlugins()
{
    impl->loadAllPlugins();
}


void PluginManagerImpl::loadAllPlugins()
{
    size_t totalNumActivated = 0;
    while(true){
        size_t numActivated = 0;
        for(size_t i=0; i < pluginInfoList.size(); ++i){
            int status = pluginInfoList[i]->status;
            if(status == PluginManager::NOT_LOADED || status == PluginManager::LOADED){
                if(loadPlugin(i)){
                    numActivated++;
                    totalNumActivated++;
                }
            }
        }
        if(numActivated == 0 || totalNumActivated == pluginInfoList.size()){
            break;
        }
    }
}


bool PluginManager::loadPlugin(int index)
{
    return impl->loadPlugin(index);
}


bool PluginManagerImpl::loadPlugin(int index)
{
    QString errorMessage;

    PluginInfoPtr info = pluginInfoList[index];
    
    if(info->status == PluginManager::ACTIVE){
        mv->putln(tr("Plugin file \"%1\" has already been activated.").arg(info->pathString.c_str()));;

    } else if(info->status == PluginManager::NOT_LOADED){
        mv->putln(tr("Loading plugin file \"%1\"").arg(info->pathString.c_str()));
        mv->flush();

        info->dll.setFileName(info->pathString.c_str());

        if(!(info->dll.load())){
            errorMessage = info->dll.errorString();

        } else {
            void* symbol = info->dll.resolve("getChoreonoidPlugin");
            if(!symbol){
                info->status = PluginManager::INVALID;
                errorMessage = tr("The plugin entry function \"getChoreonoidPlugin\" is not found.\n");
                errorMessage += info->dll.errorString();

            } else {
                Plugin::PluginEntry getCnoidPluginFunc = (Plugin::PluginEntry)(symbol);
                info->plugin = getCnoidPluginFunc();

                if(!info->plugin){
                    info->status = PluginManager::INVALID;
                    errorMessage = tr("The plugin object cannot be created.");

                } else {

                    info->status = PluginManager::LOADED;
                    info->name = info->plugin->name();

                    int numDependencies = info->plugin->numDependencies();
                    for(int i=0; i < numDependencies; ++i){
                        info->dependencies.push_back(info->plugin->dependency(i));
                    }
                }
            }
        }
    }

    if(info->status == PluginManager::LOADED){

        PluginMap::iterator p = nameToActivePluginInfoMap.find(info->name);

        if(p != nameToActivePluginInfoMap.end()){
            info->status = PluginManager::CONFLICT;
            PluginInfoPtr loadedPlugin = p->second;
            errorMessage = tr("The plugin conflicts with plugin \"%1\".").arg(loadedPlugin->pathString.c_str());

        } else {
            bool dependenciesActive = true;
                        
            std::list<std::string>::iterator p = info->dependencies.begin();
            while(p != info->dependencies.end()){
                PluginMap::iterator q = nameToActivePluginInfoMap.find(*p);
                if(q == nameToActivePluginInfoMap.end()){
                    dependenciesActive = false;
                    break;
                }
                p = info->dependencies.erase(p);
            }

            if(dependenciesActive){
                if(!info->plugin->initialize()){
                    info->status = PluginManager::INVALID;
                    errorMessage = tr("The plugin object cannot be intialized.");
                } else {
                    info->status = PluginManager::ACTIVE;
                    nameToActivePluginInfoMap[info->name] = info;

                    // set an about dialog
                    info->plugin->menuManager().setPath(tr("/Help/About Plugins"))
                        .addItem(tr("About %1 Plugin").arg(info->name.c_str()))
                        ->sigTriggered().connect(
                            bind(&PluginManagerImpl::onAboutDialogTriggered, this, info.get()));


                    // register old names
                    int numOldNames = info->plugin->numOldNames();
                    for(int i=0; i < numOldNames; ++i){
                        oldNameToCurrentPluginNameMap.insert(
                            make_pair(info->plugin->oldName(i), info->name));
                    }
                    
                    mv->putln(tr("%1-plugin has been activated.").arg(info->name.c_str()));
                    mv->flush();
                    ExtensionManager::notifySystemUpdate();
                }
            }
        }
    }

    if(!errorMessage.isEmpty()){
        mv->putln(tr("Loading the plugin failed."));
        mv->putln(errorMessage);
        mv->flush();
    }

    return (info->status == PluginManager::ACTIVE);
}


void PluginManagerImpl::onAboutDialogTriggered(PluginInfo* info)
{
    if(!info->aboutDialog){
        info->aboutDialog = new DescriptionDialog();
        info->aboutDialog->setWindowTitle(tr("About %1 Plugin").arg(info->name.c_str()));
        info->aboutDialog->setDescription(info->plugin->description());
    }

    info->aboutDialog->show();
}


const char* PluginManager::guessActualPluginName(const std::string& name)
{
    return impl->guessActualPluginName(name);
}


const char* PluginManagerImpl::guessActualPluginName(const std::string& name)
{
    PluginMap::iterator p = nameToActivePluginInfoMap.find(name);
    if(p != nameToActivePluginInfoMap.end()){
        return p->second->name.c_str();
    }

    MultiNameMap::iterator q, upper_bound;
    std::pair<MultiNameMap::iterator, MultiNameMap::iterator> range =
        oldNameToCurrentPluginNameMap.equal_range(name);
    for(MultiNameMap::iterator q = range.first; q != range.second; ++q){
        const string& candidate = q->second;
        PluginMap::iterator r = nameToActivePluginInfoMap.find(candidate);
        if(r != nameToActivePluginInfoMap.end()){
            return r->second->name.c_str();
        }
    }

    return 0;
}


bool PluginManager::unloadPlugin(int index)
{
    return impl->unloadPlugin(index);
}


bool PluginManagerImpl::unloadPlugin(int index)
{
    PluginInfoPtr info = pluginInfoList[index];

    if(info->status == PluginManager::ACTIVE){

        if(info->plugin){

            bool finalized = info->plugin->finalize();

            if(finalized){
                info->status = PluginManager::LOADED;

            } else {
                mv->putln(tr("Plugin %1% was not successfully finalized.").arg(info->name.c_str()));
                mv->putln(tr("The system might not work correctly."));
                mv->flush();
            }
        }
    }

    if(info->status == PluginManager::LOADED){

        delete info->plugin;
        info->plugin = 0;
        if(info->aboutDialog){
            delete info->aboutDialog;
            info->aboutDialog = 0;
        }
        nameToActivePluginInfoMap.erase(info->name);
        info->status = PluginManager::NOT_LOADED;

        ExtensionManager::notifySystemUpdate();

        pluginsToUnload.push_back(index);

        unloadingRequest.request();
    }
    
    return (info->status == PluginManager::NOT_LOADED);
}


void PluginManagerImpl::onUnloadingRequest()
{
    for(size_t i=0; i < pluginsToUnload.size(); ++i){
        PluginInfoPtr info = pluginInfoList[pluginsToUnload[i]];
        info->dll.unload();
        mv->putln(tr("Plugin dll %1% has been unloaded.").arg(info->pathString.c_str()));
        mv->flush();
    }
    pluginsToUnload.clear();
}


void PluginManager::unloadAllPlugins()
{
    impl->unloadAllPlugins();
}


/**
   \todo consider dependency and unloading order
*/
void PluginManagerImpl::unloadAllPlugins()
{
    for(size_t i=0; i < pluginInfoList.size(); ++i){
        unloadPlugin(i);
    }
}
